package ru.landgrafhomyak.db.serdha0.api.queries

import ru.landgrafhomyak.db.serdha0.api.misc.Column
import ru.landgrafhomyak.db.serdha0.api.table.Table
import ru.landgrafhomyak.db.serdha0.api.misc.DatabaseType
import ru.landgrafhomyak.db.serdha0.api.misc.QueryParam
import ru.landgrafhomyak.db.serdha0.api.misc._Selectable
import ru.landgrafhomyak.db.serdha0.api.runtime.InputRow
import ru.landgrafhomyak.db.serdha0.api.table.UniqueIndex

public interface InsertQuery<out qUE : Any> : _Query<qUE> {
	@Suppress("ClassName")
	public interface _UpsertClauseScope<qUE : Any, tUE : Any> {
		public fun keepColumn(column: Column<*, *, tUE>)
		public fun overrideColumn(column: Column<*, *, tUE>)
		public fun <RT, DT : DatabaseType<RT>> setFromParam(dst: Column<RT, DT, qUE>, src: QueryParam<RT, DT, qUE>)
		public fun <RT> setConstant(dst: Column<RT, *, qUE>, src: RT)
	}

	@Suppress("ClassName")
	public interface _CommonInsertCreatorScope<qUE : Any, ttUE : Any> {
		public val types: DatabaseType.Provider

		public fun onConflictUpdate(u: UniqueIndex<ttUE>, c: (_UpsertClauseScope<qUE, ttUE>) -> Unit)
		public fun onConflictThrow(u: UniqueIndex<ttUE>)
		public fun onConflictSkip(u: UniqueIndex<ttUE>)
	}


	public interface SingleRow<out qUE : Any> : InsertQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Constructor<qUE : Any, ttUE : Any> : _Query.Constructor<SingleRow<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): SingleRow<qUE> =
				outQueries.insertSingleRow(this)

			public val targetTable: Table<ttUE, *>

			public fun createInsertSingleRowWithReturning(context: Scope<qUE, ttUE>): qUE

			public interface Scope<qUE : Any, ttUE : Any> : _CommonInsertCreatorScope<qUE, ttUE> {
				public val tt: Table<ttUE, *>

				public fun <RT, DT : DatabaseType<RT>> queryParam(name: String, type: DT): QueryParam<RT, DT, qUE>

				public fun <RT, DT : DatabaseType<RT>> insertConstant(dst: Column<RT, DT, ttUE>, src: RT)
				public fun <RT, DT : DatabaseType<RT>> insertParam(dst: Column<RT, DT, ttUE>, src: QueryParam<RT, DT, qUE>)

				public fun onConflictReturnOnly(u: UniqueIndex<ttUE>)

				public fun <RT, DT : DatabaseType<RT>> returnInsertedValue(name: String, c: Column<RT, DT, ttUE>): Column<RT, DT, qUE>
				public fun <RT, DT : DatabaseType<RT>> returnParam(name: String, c: QueryParam<RT, DT, ttUE>): Column<RT, DT, qUE>
			}
		}
	}


	public interface FromSubquery<out qUE : Any> : InsertQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Constructor<qUE : Any, ttUE : Any, sUE : Any> : _Query.Constructor<FromSubquery<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): FromSubquery<qUE> =
				outQueries.insertFromSubquery(this)

			public val targetTable: Table<ttUE, *>
			public val selector: _Selectable<sUE>

			public fun createInsertSingleRowWithReturning(context: Scope<qUE, ttUE, sUE>): qUE

			public interface Scope<qUE : Any, ttUE : Any, sUE : Any> : _CommonInsertCreatorScope<qUE, ttUE> {
				public val tt: Table<ttUE, *>
				public val src: _Selectable<sUE>

				public fun <RT, DT : DatabaseType<RT>> queryParam(name: String, type: DT): QueryParam<RT, DT, qUE>

				public val sourceQueryParams: InputRow.WithRedirect<sUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> insertConstant(dst: Column<RT, DT, ttUE>, src: RT)
				public fun <RT, DT : DatabaseType<RT>> insertParam(dst: Column<RT, DT, ttUE>, src: QueryParam<RT, DT, qUE>)
				public fun <RT, DT : DatabaseType<RT>> insertFromSubquery(dst: Column<RT, DT, ttUE>, src: Column<RT, DT, ttUE>)

				public fun onConflictReturnOnly(u: UniqueIndex<ttUE>)

				public fun <RT, DT : DatabaseType<RT>> returnFromSubquery(name: String, column: Column<RT, DT, sUE>): Column<RT, DT, qUE>
				public fun <RT, DT : DatabaseType<RT>> returnNewValue(name: String, column: Column<RT, DT, ttUE>): Column<RT, DT, qUE>
				public fun <RT, DT : DatabaseType<RT>> returnParam(name: String, param: QueryParam<RT, DT, ttUE>): Column<RT, DT, qUE>
			}
		}
	}

	public interface MultipleRows<out qUE : Any> : InsertQuery<qUE>, _Query.Table2Void<qUE> {
		public interface Constructor<qUE : Any, ttUE : Any> : _Query.Constructor<MultipleRows<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): MultipleRows<qUE> =
				outQueries.insertMultipleRows(this)

			public val targetTable: Table<ttUE, *>
			public fun createInsertMultipleRows(context: Scope<qUE, ttUE>): qUE

			public interface Scope<qUE : Any, ttUE : Any> : _CommonInsertCreatorScope<qUE, ttUE> {
				public val tt: Table<ttUE, *>

				public fun <RT, DT : DatabaseType<RT>> inputColumn(name: String, type: DT): QueryParam<RT, DT, qUE>

				public fun <RT, DT : DatabaseType<RT>> insertColumn(dst: Column<RT, DT, ttUE>, src: QueryParam<RT, DT, qUE>)
			}
		}
	}
}