package ru.landgrafhomyak.db.serdha0.api.queries

import ru.landgrafhomyak.db.serdha0.api.misc.Column
import ru.landgrafhomyak.db.serdha0.api.misc.IntermediateColumn
import ru.landgrafhomyak.db.serdha0.api.misc.DatabaseType
import ru.landgrafhomyak.db.serdha0.api.misc.Mapper
import ru.landgrafhomyak.db.serdha0.api.misc.FiltersScope
import ru.landgrafhomyak.db.serdha0.api.misc.QueryParam
import ru.landgrafhomyak.db.serdha0.api.misc.Reducer
import ru.landgrafhomyak.db.serdha0.api.misc._Selectable
import ru.landgrafhomyak.db.serdha0.api.runtime.InputRow

public interface SelectQuery<out qUE : Any> : _Query<qUE> {
	public enum class Order {
		ASC, DESC
	}

	public class JoinType(
		public val leftNull: LeftNull,
		public val join: Join
	) {
		public enum class LeftNull {
			AS_ID, SKIP, JOIN_TO_NULL
		}

		public enum class Join {
			LEFT_TO_FIRST_RIGHT, LEFT_FOR_EACH_RIGHT, LEFT_FOR_EACH_RIGHT_OR_NULL,
		}
	}

	@Suppress("ClassName")
	public interface _CommonSelectCreatorScope<qUE : Any> {
		public val types: DatabaseType.Provider
		public fun <RT, DT : DatabaseType<RT>> queryParam(name: String, type: DT): QueryParam<RT, DT, qUE>
		public fun <RT, DT : DatabaseType<RT>> paramAsColumn(param: QueryParam<RT, DT, qUE>): IntermediateColumn<RT, DT, qUE>

		public fun <mUE : Any> createMapper(constructor: Mapper.Constructor<mUE>): Mapper<mUE>
		public fun <mUE : Any, oUE : Any> mapColumns(
			mapper: Mapper<mUE>,
			input: Mapper.InputLinker<mUE, qUE>,
			output: Mapper.OutputLinker<mUE, qUE, oUE>
		): oUE

		public val filters: FiltersScope<qUE>
		public fun orderBy(order: Order, vararg column: IntermediateColumn<*, *, qUE>)

		public var limit: UInt
		public var offset: ULong

		public fun <RT, DT : DatabaseType<RT>> returnParam(name: String, param: QueryParam<RT, DT, qUE>): Column<RT, DT, qUE>
		public fun <RT, DT : DatabaseType<RT>> returnColumn(name: String, column: IntermediateColumn<RT, DT, qUE>): Column<RT, DT, qUE>
	}

	public interface Simple<out qUE : Any> : SelectQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Constructor<qUE : Any, tUE : Any> : _Query.Constructor<Simple<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): Simple<qUE> =
				outQueries.select(this)

			public val source: _Selectable<tUE>

			public fun createSelect(context: Scope<qUE, tUE>): qUE

			public interface Scope<qUE : Any, tUE : Any> : _CommonSelectCreatorScope<qUE> {
				public val src: _Selectable<tUE>
				public val subqueryParams: InputRow.WithRedirect<tUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> selectColumnFromSubquery(param: Column<RT, DT, tUE>): IntermediateColumn<RT, DT, qUE>
			}
		}
	}

	public interface WithJoin<out qUE : Any> : SelectQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Constructor<qUE : Any, lsqUE : Any, rsqUE : Any> : _Query.Constructor<WithJoin<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): WithJoin<qUE> =
				outQueries.selectWithJoin(this)

			public val left: _Selectable<lsqUE>
			public val right: _Selectable<rsqUE>
			public val joinType: JoinType
			public val referenceColumn: Column<*, *, lsqUE>
			public val targetColumn: Column<*, *, rsqUE>

			public fun createSelectWithJoin(context: Scope<qUE, lsqUE, rsqUE>): qUE

			public interface Scope<qUE : Any, lsqUE : Any, rsqUE : Any> : _CommonSelectCreatorScope<qUE> {
				public val lsq: _Selectable<lsqUE>
				public val rsq: _Selectable<rsqUE>

				public val leftSubqueryParams: InputRow.WithRedirect<lsqUE, qUE>
				public val rightSubqueryParams: InputRow.WithRedirect<rsqUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> selectColumnFromLeft(column: Column<RT, DT, lsqUE>): IntermediateColumn<RT, DT, qUE>
				public fun <RT, DT : DatabaseType<RT>> selectColumnFromRight(column: Column<RT, DT, rsqUE>): IntermediateColumn<RT, DT, qUE>
				public fun <RT : Any, DT : DatabaseType<RT>> selectColumnFromRightAsNullable(column: Column<RT, DT, rsqUE>): IntermediateColumn<RT?, DatabaseType.Nullable<RT, DT>, qUE>
			}
		}
	}

	public interface Reducing<out qUE : Any> : SelectQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Constructor<qUE : Any, tUE : Any> : _Query.Constructor<Reducing<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): Reducing<qUE> =
				outQueries.selectReducing(this)

			public val source: _Selectable<tUE>

			public fun createSelectWithReduce(context: Scope<qUE, tUE>): qUE

			public interface Scope<qUE : Any, tUE : Any> : _CommonSelectCreatorScope<qUE> {
				public val src: _Selectable<tUE>
				public val sourceParams: InputRow.WithRedirect<tUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> selectColumnFromSubquery(param: Column<RT, DT, tUE>): IntermediateColumn<RT, DT, qUE>

				public fun <rUE : Any> createReducer(constructor: Reducer.Constructor<rUE>): Reducer<rUE>
				public fun <mUE : Any, oUE : Any> reduceRows(
					reducer: Reducer<mUE>,
					linker: Reducer.Linker<mUE, qUE, oUE>,
				): oUE

				public fun groupBy(vararg columns: IntermediateColumn<*, *, qUE>)
			}
		}
	}

	public interface Folding<out qUE : Any> : SelectQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Constructor<qUE : Any, tUE : Any> : _Query.Constructor<Folding<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): Folding<qUE> =
				outQueries.selectFolding(this)

			public val source: _Selectable<tUE>

			public fun createSelectWithFold(context: Scope<qUE, tUE>): qUE

			public interface Scope<qUE : Any, tUE : Any> : _CommonSelectCreatorScope<qUE> {
				public val src: _Selectable<tUE>
				public val sourceParams: InputRow.WithRedirect<tUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> selectColumnFromSubquery(param: Column<RT, DT, tUE>): IntermediateColumn<RT, DT, qUE>

				public fun <rUE : Any> createReducer(constructor: Reducer.Constructor<rUE>): Reducer<rUE>
				public fun <mUE : Any, oUE : Any> foldRows(
					reducer: Reducer<mUE>,
					linker: Reducer.Linker<mUE, qUE, oUE>,
				): oUE

				public fun groupBy(vararg columns: IntermediateColumn<*, *, qUE>)
			}
		}
	}


	public interface Union<out qUE : Any> : SelectQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Constructor<qUE : Any, fqUE : Any, lqUE : Any> : _Query.Constructor<Union<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): Union<qUE> =
				outQueries.selectUnion(this)

			public val first: _Selectable<fqUE>
			public val last: _Selectable<lqUE>

			public fun createSelectWithJoin(context: Scope<qUE, fqUE, lqUE>): qUE

			public interface Scope<qUE : Any, fqUE : Any, lqUE : Any> : _CommonSelectCreatorScope<qUE> {
				public val fq: _Selectable<fqUE>
				public val lq: _Selectable<lqUE>

				public val firstSubqueryParams: InputRow.WithRedirect<fqUE, qUE>
				public val lastSubqueryParams: InputRow.WithRedirect<lqUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> uniteColumns(first: Column<RT, DT, fqUE>, last: Column<RT, DT, lqUE>): IntermediateColumn<RT, DT, qUE>
			}
		}
	}
}